import logging
import multiprocessing
import typing as t
from dataclasses import dataclass, field

from torch import nn, Tensor
from torch.optim import Optimizer
from torch.utils.data import default_collate

ModelOutputType = t.Union[Tensor, list, tuple]
"""Type representing a model output artifact. Can be a tensor, a nested list or nested tuple of tensors."""

ModelOutputsType = dict[str, ModelOutputType]
"""Type representing multiple model output artifacts: a dictionary with artifact names as keys and model output artifacts as values."""

DataType = t.Union[dict[str, t.Any], t.Iterable]
"""Type representing model input data. Can be a nested object of dictionaries, lists, sets, tuples containing arbitrary types such as tensors or strings."""

TrackerConfigType = dict[str, t.Union[float, int, bool, str, None]]
"""Type representing dictionary with parameters logged to experiment tracker"""


class LRSchedulerProtocol(t.Protocol):
    """
    A protocol (https://peps.python.org/pep-0544/) representing a learning rate scheduler.
    """

    def step(self):
        """Function that updates a scheduler each training step."""
        ...

    def state_dict(self):
        """Function that returns scheduler's internal state dict."""
        ...

    def load_state_dict(self, state_dict):
        """Function that loads this scheduler's state from a state dict."""
        ...


@dataclass
class XZTrainerConfig:
    """
    A configuration for XZTrainer training setup
    """

    experiment_name: str
    """Name of current experiment. Used for logging and checkpoint saving"""

    minibatch_size: int
    """
    Minibatch size used for training.
    In case of a multi-gpu or gradient accumulated train setup, it is a physical batch size for a single GPU.
    """

    minibatch_size_eval: int
    """
    Minibatch size used for evaluation.
    In case of a multi-gpu or gradient accumulated train setup, it is a physical batch size for a single GPU.
    """

    epochs: int
    """
    Number of epochs to train for.
    """

    optimizer: t.Callable[[nn.Module], Optimizer]
    """
    A callable that creates an Optimizer object from a Module object.
    Module object can be used to pass its parameters to Optimizer constructor, for example.
    
    Examples:
        >>> optimizer=lambda module: AdamW(module.parameters(), lr=1e-3, weight_decay=1e-4)
    """

    scheduler: t.Callable[[Optimizer, int], LRSchedulerProtocol]
    """
    A callable that creates a Learning Rate Scheduler object from an Optimizer and total number of training steps.
    
    Examples:
        >>> scheduler=lambda optim, total_steps: transformers.get_linear_schedule_with_warmup(optim, int(0.1 * total_steps), total_steps)
    """

    gradient_clipping: float = 1.0
    """
    Max gradient norm used for gradient clipping. Can be set to `0` to disable gradient clipping.
    """

    dataloader_num_workers: int = multiprocessing.cpu_count()
    """
    Number of worker processes used for data loaders.
    """

    dataloader_pin_memory: bool = True
    """
    Boolean value indicating whether to pin memory for data loaders.
    """

    dataloader_persistent_workers: bool = True
    """
    Boolean value indicating whether to keep workers alive after each epoch.
    """

    dataloader_shuffle_train_dataset: bool = True
    """
    Boolean value indicating whether to shuffle a training dataset within a dataloader.
    """

    log_steps: int = 100
    """
    Controls frequency of logging model metrics.
    
    Can be either:
    
    - `-1` - to disable logging
    
    - `0` - to log only after an epoch finishes
    
    - Any positive value - to log at each Nth training step
    """

    eval_steps: int = 0
    """
    Controls frequency of running a model on evaluation set.

    Can be either:

    - `-1` - to disable evaluation

    - `0` - to evaluate only after an epoch finishes

    - Any positive value - to evaluate at each Nth training step
    """

    skip_nan_loss: bool = True
    """
    Boolean value indicating whether to skip update when a NaN loss value occurs.
    """

    save_steps: int = 100
    """
    Controls frequency of saving a model and trainer state.

    Can be either:

    - `-1` - to disable saving

    - `0` - to save only after an epoch finishes

    - Any positive value - to save at each Nth training step
    """

    save_keep_n: int = 3
    """
    Number of saves to keep while training. After a new save finished, all the other saves will be deleted except for `save_keep_n` last ones. Set to `-1` to disable save removal.
    """

    collate_fn: t.Callable[[list[object]], t.Any] = default_collate
    """
    A function used to collate multiple dataset items into a single batch.
    """

    tracker_config: dict[str, t.Any] = field(default_factory=dict)
    """
    Arbitrary hyperparameters logged to experiment tracker.
    """

    tracker_init_kwargs: dict[str, t.Any] = field(default_factory=dict)
    """
    Arbitrary keyword arguments used for Accelerate experiment tracker. Directly passed to `accelerator.init_trackers(...)`. See [Accelerate docs](https://huggingface.co/docs/accelerate/en/usage_guides/tracking)
    """

    tracker_logging_dir: t.Optional[str] = None
    """
    Logging directory (or remote address) for Accelerate experiment tracker. If `None` - it will be set to `{project_dir}/runs` automatically.
    """

    logging_level: t.Union[int, None] = logging.INFO
    """
    Logging level that xztrainer initializes Python loggers with. Set to `None` to make xztrainer not do anything with Python loggers.
    """


@t.runtime_checkable
class MetricMultiOutputNamedProtocol(t.Protocol):
    @property
    def multi_output_names(self) -> t.List[str]:
        ...



import torch
from accelerate.utils import set_seed


def set_seeds(seed: int):
    """
    Sets all the RNG states (python, torch, numpy). Internally calls accelerate's set_seed, left only for backward compatibility.

    Args:
        seed: RNG state
    """
    set_seed(seed)


def enable_tf32():
    """
    Enables tensorfloat32 computations for pytorch.
    """
    torch.backends.cuda.matmul.allow_tf32 = True
    torch.backends.cudnn.allow_tf32 = True



from collections import OrderedDict
from typing import Mapping, Any

from accelerate import Accelerator
from torch import Tensor

from xztrainer.model import ModelOutputType, DataType


def detach_tensor(output: ModelOutputType, move_to_cpu: bool) -> ModelOutputType:
    """
    Recursively detaches each tensor in a model output structure.

    Args:
        output: Model output to detach tensors in
        move_to_cpu: Whether to move a tensor to CPU after detaching

    Returns:

    """
    if isinstance(output, Tensor):
        output = output.detach()
        if move_to_cpu:
            output = output.cpu()
        return output
    elif isinstance(output, list):
        return [detach_tensor(x, move_to_cpu) for x in output]
    elif isinstance(output, tuple):
        return tuple(detach_tensor(x, move_to_cpu) for x in output)
    else:
        return output


def move_data_to_device(data: Any, accelerator: Accelerator) -> DataType:
    """
    Recursively moves each tensor in a data structure to device specified by `Accelerator` object.
    Args:
        data: Structure to move tensors in
        accelerator: `Accelerator` object

    Returns:

    """
    if isinstance(data, Tensor):
        return data.to(accelerator.device)
    elif isinstance(data, OrderedDict):
        return OrderedDict({k: move_data_to_device(v, accelerator) for k, v in data.items()})
    elif isinstance(data, Mapping):
        return {k: move_data_to_device(v, accelerator) for k, v in data.items()}
    elif isinstance(data, tuple):
        return tuple(move_data_to_device(v, accelerator) for v in data)
    elif isinstance(data, list):
        return [move_data_to_device(v, accelerator) for v in data]
    elif isinstance(data, set):
        return set(move_data_to_device(v, accelerator) for v in data)
    else:
        return data



import typing as t

from accelerate import Accelerator
from accelerate.optimizer import AcceleratedOptimizer
from accelerate.scheduler import AcceleratedScheduler
from torch.optim import Optimizer
from torch.utils.data import DataLoader

from xztrainer.context import TrainContext
from xztrainer.model import XZTrainerConfig, TrackerConfigType, LRSchedulerProtocol


def _prepend_tracker_config(config: TrackerConfigType, prefix: str) -> TrackerConfigType:
    return {f'{prefix}_{k}': v for k, v in config.items()}


def _filter_dict_for_tracker_confix(values: dict[str, t.Any]) -> TrackerConfigType:
    out_dict = {}

    for k, v in values.items():
        if isinstance(v, (float, int, bool, str)) or v is None:
            out_dict[k] = v
        elif isinstance(v, (tuple, str)):
            out_dict.update(_filter_dict_for_tracker_confix({f'{k}_{i}': itm for i, itm in enumerate(v)}))

    return out_dict


def _extract_optimizer_config(optimizer: Optimizer) -> TrackerConfigType:
    if isinstance(optimizer, AcceleratedOptimizer):
        name = optimizer.optimizer.__class__.__name__
    else:
        name = optimizer.__class__.__name__
    dct = {
        'name': name,
        **_filter_dict_for_tracker_confix(optimizer.defaults)
    }
    dct.pop('foreach', None)
    dct.pop('fused', None)
    dct.pop('differentiable', None)
    dct.pop('capturable', None)
    dct.pop('maximize', None)
    return dct


def _extract_accelerator_config(accelerator: Accelerator) -> TrackerConfigType:
    return {
        'gradient_accumulation_steps': accelerator.gradient_accumulation_steps,
        'num_processes': accelerator.num_processes,
        'mixed_precision': accelerator.mixed_precision
    }


def _extract_trainer_config(config: XZTrainerConfig) -> TrackerConfigType:
    return {
        'minibatch_size': config.minibatch_size,
        'epochs': config.epochs,
        'gradient_clipping': config.gradient_clipping,
        'shuffle_train_dataset': config.dataloader_shuffle_train_dataset,
        'skip_nan_loss': config.skip_nan_loss
    }


def _extract_dataloader_config(dataloader: DataLoader) -> TrackerConfigType:
    return {
        'len': len(dataloader)
    }

def _extract_lr_scheduler_config(scheduler: LRSchedulerProtocol) -> TrackerConfigType:
    if isinstance(scheduler, AcceleratedScheduler):
        name = scheduler.scheduler.__class__.__name__
    else:
        name = scheduler.__class__.__name__
    return {
        'name': name
    }

def create_tracker_config(context: TrainContext) -> TrackerConfigType:
    out_dict = {
        **_prepend_tracker_config(context.trainer.config.tracker_config, 'config'),
        **_prepend_tracker_config(context.trainer.trainable.tracker_config(context), 'trainable'),
        **_prepend_tracker_config(_extract_trainer_config(context.trainer.config), 'trainer'),
        **_prepend_tracker_config(_extract_optimizer_config(context.optimizer), 'optimizer'),
        **_prepend_tracker_config(_extract_accelerator_config(context.trainer.accelerator), 'accelerator'),
        **_prepend_tracker_config(_extract_dataloader_config(context.data_loader), 'dataloader'),
        **_prepend_tracker_config(_extract_lr_scheduler_config(context.scheduler), 'scheduler')
    }
    return out_dict



import math
import re
import shutil
import typing as t
from logging import basicConfig
from pathlib import Path

import torch
from accelerate import Accelerator
from accelerate.logging import get_logger
from accelerate.utils.tqdm import tqdm
from torch import nn, Tensor
from torch.utils.data import Dataset, DataLoader
from torchmetrics import Metric

from xztrainer.auto_tracker_config import create_tracker_config
from xztrainer.context import ContextType, TrainContext, BaseContext, BaseTrainContext, EvalContext
from xztrainer.functional import count_parameters
from xztrainer.logger import AccelerateLogger
from xztrainer.model import XZTrainerConfig, DataType, ModelOutputsType, MetricMultiOutputNamedProtocol
from xztrainer.tensor_helper import detach_tensor, move_data_to_device
from xztrainer.trainable import XZTrainable

_RE_SAVE_NAME = re.compile(r'save-\d+')


class XZTrainState:
    def __init__(self, accelearator: Accelerator, print_metrics: dict[str, Metric], train_metrics: dict[str, Metric],
                 eval_metrics: dict[str, Metric]):
        self.accelerator = accelearator
        self.print_metrics = print_metrics
        self.train_metrics = train_metrics
        self.eval_metrics = eval_metrics
        self.current_epoch = 1
        self.passed_steps = 0

    @staticmethod
    def _synced_state_dict(metric: Metric):
        with metric.sync_context():
            return metric.state_dict()

    @staticmethod
    def _metrics_to_state_dict(metrics: dict[str, Metric]) -> dict[str, dict[str, t.Any]]:
        return {k: XZTrainState._synced_state_dict(v) for k, v in metrics.items()}

    def _load_metrics_from_state_dict(self, metrics: dict[str, Metric], state_dict: dict[str, dict[str, t.Any]]):
        for k, v in metrics.items():
            v.load_state_dict(move_data_to_device(state_dict[k], self.accelerator))

    def update_epoch(self):
        self.current_epoch += 1

    def update_passed_steps(self):
        self.passed_steps += 1

    def load_state_dict(self, dct: dict[str, t.Any]):
        self._load_metrics_from_state_dict(self.print_metrics, dct['print_metrics'])
        self._load_metrics_from_state_dict(self.train_metrics, dct['train_metrics'])
        self._load_metrics_from_state_dict(self.eval_metrics, dct['eval_metrics'])
        self.current_epoch = dct['current_epoch']
        self.passed_steps = dct['passed_steps']

    def state_dict(self) -> dict[str, t.Any]:
        return {
            'print_metrics': self._metrics_to_state_dict(self.print_metrics),
            'train_metrics': self._metrics_to_state_dict(self.train_metrics),
            'eval_metrics': self._metrics_to_state_dict(self.eval_metrics),
            'current_epoch': self.current_epoch,
            'passed_steps': self.passed_steps
        }


class XZTrainer:
    config: XZTrainerConfig
    """Configuration used by this trainer"""

    def __init__(
            self,
            config: XZTrainerConfig,
            model: nn.Module,
            trainable: XZTrainable,
            accelerator: Accelerator
    ):
        """
        Creates a trainer instance

        Args:
            config: Trainer config to use
            model: Model instance to train
            trainable: Trainable instance to use
            accelerator: Accelerator object
        """
        self.config = config

        self.accelerator = accelerator
        self.model = model
        self.trainable = trainable
        self.logger = get_logger(name='xztrainer')

    def _create_dataloader(self, data: Dataset, **kwargs) -> DataLoader:
        return DataLoader(
            data,
            collate_fn=self.config.collate_fn,
            num_workers=self.config.dataloader_num_workers,
            persistent_workers=self.config.dataloader_persistent_workers,
            pin_memory=self.config.dataloader_pin_memory,
            **kwargs
        )

    def _create_metrics(self, context_type: ContextType) -> dict[str, Metric]:
        metrics = {}
        for k, metric in self.trainable.create_metrics(context_type).items():
            metric.persistent(True)
            metrics[k] = metric.to(self.accelerator.device)
        return metrics

    @staticmethod
    def _set_training_state(context: BaseContext):
        context.model.train()
        if isinstance(context, BaseTrainContext):
            context.logger.update_top_classifier(('step', 'train'))

    @staticmethod
    def _set_evaluating_state(context: BaseContext):
        context.model.eval()
        if isinstance(context, BaseTrainContext):
            context.logger.update_top_classifier(('step', 'eval'))

    def _forward_pass(self, context: BaseContext, data: DataType) -> tuple[Tensor, ModelOutputsType]:
        loss, model_output = self.trainable.step(context, data)
        return loss, model_output

    def _update_metrics(self, context_type: ContextType, model_outputs: ModelOutputsType, metrics: dict[str, Metric]):
        model_outputs = {k: detach_tensor(v, move_to_cpu=False) for k, v in model_outputs.items()}
        self.trainable.update_metrics(context_type, model_outputs, metrics)

    def _calculate_reset_metrics(
            self, trainable: XZTrainable, context_type: ContextType, metrics: dict[str, Metric]
    ) -> dict[str, float]:
        metric_values = {}
        for name, metric in metrics.items():
            metric_val = metric.compute()
            metric_val_els = metric_val.numel()
            if metric_val_els == 0:
                raise ValueError(f'empty metric {name}')
            elif metric_val_els == 1:
                metric_values[name] = metric_val.item()
            else:
                if isinstance(metric, MetricMultiOutputNamedProtocol):
                    metric_names = metric.multi_output_names
                else:
                    metric_names = [str(i) for i in range(metric_val_els)]
                for itm_name, itm in zip(metric_names, metric_val.flatten()):
                    metric_values[f'{name}_{itm_name}'] = itm.item()
            metric.reset()
        metric_values.update(trainable.calculate_composition_metrics(context_type, metric_values))
        return metric_values

    def _log_trainable(self, context: BaseTrainContext, metrics: dict[str, Metric]):
        self.accelerator.wait_for_everyone()  # wait for everyone to contribute to metrics
        for k, v in self._calculate_reset_metrics(self.trainable, context.context_type, metrics).items():
            context.logger.log_scalar(k, v)
        self.trainable.log(context)

    def _cleanup_saves(self):
        if self.config.save_keep_n >= 0:
            save_dir = Path(self.accelerator.project_dir) / 'checkpoint' / self.config.experiment_name
            save_files = sorted(self._get_save_files(save_dir), reverse=True)
            save_files_to_delete = save_files[self.config.save_keep_n:]
            for step, file in save_files_to_delete:
                shutil.rmtree(file)

    def _save(self, step: int):
        save_dir = Path(self.accelerator.project_dir) / 'checkpoint' / self.config.experiment_name
        save_dir.mkdir(exist_ok=True, parents=True)
        save_path = save_dir / f'save-{step}'
        self.accelerator.wait_for_everyone()
        self.accelerator.save_state(str(save_path))
        self.accelerator.wait_for_everyone()
        if self.accelerator.is_main_process:
            self._cleanup_saves()

    def _load(self):
        save_dir = Path(self.accelerator.project_dir) / 'checkpoint' / self.config.experiment_name
        if save_dir.is_dir():
            save_files = self._get_save_files(save_dir)
            if len(save_files) > 0:
                save_file = max(save_files)[1]
                self.logger.info(f'Loading state from {save_file}')
                return self.accelerator.load_state(str(save_file))

    @staticmethod
    def _get_save_files(save_dir: Path) -> list[tuple[int, Path]]:
        save_files = [x for x in save_dir.iterdir() if _RE_SAVE_NAME.fullmatch(x.name) and x.is_dir()]
        save_files_with_step = [(int(x.stem.split('-')[1]), x) for x in save_files]
        return save_files_with_step

    def _train_epoch(self, context: TrainContext):
        self._set_training_state(context)
        context.progress_bar.update(
            context.train_state.passed_steps - context.progress_bar.n
        )

        for data in context.data_loader:
            current_step = context.train_state.passed_steps + 1

            if self.accelerator.sync_gradients:
                context.logger.update_time_step(current_step)

            with self.accelerator.accumulate(context.model):
                loss, model_out = self._forward_pass(context, data)
                model_out = self.trainable.cut_model_outputs(
                    context, model_out,
                    self.accelerator.gradient_state.remainder
                )

                self._update_metrics(context.context_type, model_out, context.train_state.print_metrics)
                self._update_metrics(context.context_type, model_out, context.train_state.train_metrics)

                self.accelerator.backward(loss)

                if self.accelerator.sync_gradients and self.accelerator.is_main_process:
                    for group_i, group in enumerate(context.optimizer.param_groups):
                        context.logger.log_scalar(['lr', str(group_i)], group['lr'])

                if self.accelerator.sync_gradients:
                    self.trainable.on_pre_update(context, current_step)

                if self.accelerator.sync_gradients:
                    l2_grad_norm = torch.norm(
                        torch.stack(
                            [torch.norm(p.grad.detach(), 2.0)
                             for p in context.model.parameters()
                             if p.grad is not None]
                        ),
                        2
                    ).item()
                    context.logger.log_scalar('l2 grad norm before clip', l2_grad_norm)
                    max_norm = context.trainer.config.gradient_clipping
                    if max_norm > 0:
                        self.accelerator.clip_grad_norm_(context.model.parameters(), max_norm=max_norm)

                context.optimizer.step()
                if self.accelerator.sync_gradients:
                    context.scheduler.step()
                context.optimizer.zero_grad()

                if self.accelerator.sync_gradients:
                    self.trainable.on_update(context, current_step)

            if self.accelerator.sync_gradients:
                if context.should_perform_step_action(self.config.log_steps, current_step):
                    self._log_trainable(context, context.train_state.print_metrics)

                if context.evaluate_data_loader and context.should_perform_step_action(self.config.eval_steps,
                                                                                       current_step):
                    self._set_evaluating_state(context)
                    context_eval = EvalContext.from_train_context(context)
                    with torch.no_grad():
                        for eval_data in context_eval.data_loader:
                            _, model_out_eval = self._forward_pass(context_eval, eval_data)
                            model_out_eval = self.trainable.cut_model_outputs(
                                context, model_out_eval,
                                self.accelerator.gradient_state.remainder
                            )
                            self._update_metrics(context_eval.context_type, model_out_eval,
                                                 context.train_state.eval_metrics)
                    self._log_trainable(context_eval, context.train_state.eval_metrics)
                    self._set_training_state(context)
                context.train_state.update_passed_steps()
                if self.accelerator.gradient_state.end_of_dataloader:
                    context.train_state.update_epoch()
                if context.should_perform_step_action(self.config.save_steps, current_step):
                    self._save(current_step)
                context.progress_bar.update()

        if len(context.data_loader) > 0:
            context.logger.update_top_classifier(('epoch', 'train'))
            context.logger.update_time_step(context.train_state.current_epoch)
            self._log_trainable(context, context.train_state.train_metrics)

    def train(self, train_data: Dataset, eval_data: t.Optional[Dataset]):
        """
        This function:

        1. Prepares all the objects used for further training, such as optimizer, dataloaders, metrics, learning rate scheduler, logger.
        2. Wraps these objects using `accelerator.prepare(...)`
        3. If a training checkpoint exists - loads it and initializes all the previously prepared objects from it.
        4. Prints out total parameter count used for training
        5. Trains for N epochs, where N specified in trainer config
        6. While training, the trainer can save its state, run evaluation and log metrics, - it is specified in trainer config.

        Args:
            train_data: Dataset used for training
            eval_data: Dataset used for evaluation. Can be `None` to disable evaluation.
        """
        if self.accelerator.project_configuration.project_dir is None:
            self.accelerator.project_configuration.project_dir = '.'
        if self.config.logging_level is not None:
            basicConfig(format='%(asctime)s [%(name)s] [%(levelname)s] %(message)s', level=self.config.logging_level)

        if self.config.tracker_logging_dir is None:
            log_dir = Path(self.accelerator.project_configuration.project_dir) / 'runs'
            log_dir.mkdir(parents=True, exist_ok=True)
            self.accelerator.project_configuration.logging_dir = str(log_dir)
        else:
            self.accelerator.project_configuration.logging_dir = self.config.tracker_logging_dir

        # Initialize and wrap model, optimizer and scheduler
        optim = self.config.optimizer(self.model)
        train_dl = self._create_dataloader(train_data, batch_size=self.config.minibatch_size)
        sync_steps_per_epoch_unscaled = int(math.ceil(len(train_dl) / self.accelerator.gradient_accumulation_steps))
        sync_steps_unscaled = sync_steps_per_epoch_unscaled * self.config.epochs
        scheduler = self.config.scheduler(optim, sync_steps_unscaled)
        model_unwrapped = self.model
        metrics_print = self._create_metrics(ContextType.TRAIN)
        metrics_train = self._create_metrics(ContextType.TRAIN)
        metrics_eval = self._create_metrics(ContextType.EVAL)
        if eval_data:
            eval_dl = self._create_dataloader(eval_data, batch_size=self.config.minibatch_size_eval)
        else:
            eval_dl = None

        model, optim, scheduler, train_dl, eval_dl = self.accelerator.prepare(
            model_unwrapped, optim, scheduler, train_dl, eval_dl
        )
        sync_steps_per_epoch = int(math.ceil(len(train_dl) / self.accelerator.gradient_accumulation_steps))
        sync_steps = sync_steps_per_epoch * self.config.epochs

        additional_state = XZTrainState(self.accelerator, metrics_print, metrics_train, metrics_eval)
        self.accelerator.register_for_checkpointing(additional_state)

        self._load()
        start_epoch = additional_state.current_epoch

        self.logger.info(f"Starting training experiment...", main_process_only=True)
        self.logger.info(f'Total synchronized steps: {sync_steps:,}', main_process_only=True)
        self.logger.info(f'Parameters [total]: {count_parameters(model_unwrapped):,}', main_process_only=True)
        self.logger.info(f'Parameters [train]: {count_parameters(model_unwrapped, lambda p: p.requires_grad):,}',
                         main_process_only=True)
        self.logger.info(f'Parameters [fixed]: {count_parameters(model_unwrapped, lambda p: not p.requires_grad):,}',
                         main_process_only=True)

        # Run epoch loop
        logger = AccelerateLogger(self.accelerator)
        with tqdm(total=sync_steps, desc=f'Train') as progress_bar:
            for epoch in range(start_epoch, self.config.epochs + 1):
                if epoch == start_epoch and additional_state.passed_steps != 0:  # resumed
                    steps_remainder = additional_state.passed_steps % sync_steps_per_epoch
                    passed_minibatches_last_epoch = steps_remainder * self.accelerator.gradient_accumulation_steps
                    train_dl_current = self.accelerator.skip_first_batches(
                        train_dl,
                        num_batches=passed_minibatches_last_epoch
                    )
                else:
                    train_dl_current = train_dl
                context = TrainContext(
                    trainer=self,
                    logger=logger,
                    optimizer=optim,
                    scheduler=scheduler,
                    data_loader=train_dl_current,
                    model=model,
                    model_unwrapped=model_unwrapped,
                    sync_steps=sync_steps,
                    evaluate_data_loader=eval_dl,
                    progress_bar=progress_bar,
                    train_state=additional_state
                )
                if epoch == start_epoch:
                    self.accelerator.init_trackers(
                        self.config.experiment_name,
                        config=create_tracker_config(context),
                        init_kwargs=self.config.tracker_init_kwargs
                    )
                    self.trainable.on_load(context, additional_state.passed_steps + 1)
                self._train_epoch(context)
        self.accelerator.end_training()



import abc
import typing as t
from dataclasses import dataclass
from enum import Enum

from torch import nn
from torch.optim import Optimizer
from torch.utils.data import DataLoader
from tqdm import tqdm

from xztrainer.logger import AccelerateLogger
from xztrainer.model import LRSchedulerProtocol

if t.TYPE_CHECKING:
    from xztrainer.trainer import XZTrainer, XZTrainState


class ContextType(Enum):
    """
    Context type object. Can be used for checking current environment.
    """

    TRAIN = 'train'
    """Training context"""
    EVAL = 'eval'
    """Evaluation context"""
    INFERENCE = 'inference'
    """Inference context"""

@dataclass
class BaseContext(abc.ABC):
    """Base context that contains fields shared across training, evaluation and inference context."""

    trainer: 'XZTrainer'
    """Trainer instance"""

    data_loader: DataLoader
    """DataLoader instance"""

    model: nn.Module
    """Model instance. The model is wrapped using Accelerate"""

    @property
    @abc.abstractmethod
    def context_type(self) -> ContextType:
        ...


@dataclass
class BaseTrainContext(BaseContext):
    """Base train context that contains fields shared across training and evaluation context."""

    logger: AccelerateLogger
    """Logger instance"""

    optimizer: Optimizer
    """Optimizer instance"""

    scheduler: LRSchedulerProtocol
    """Learning rate scheduler instance"""

    model_unwrapped: nn.Module
    """Model instance. The model is not wrapped using Accelerate."""

    train_state: 'XZTrainState'
    """Custom state used by trainer"""


@dataclass
class TrainContext(BaseTrainContext):
    """Context used for training"""

    sync_steps: int
    """Total count of training steps"""

    progress_bar: tqdm
    """Progress bar used for displaying progress"""

    evaluate_data_loader: t.Union[DataLoader, None]
    """Evaluation DataLoader instance"""

    @property
    def context_type(self) -> ContextType:
        return ContextType.TRAIN

    def should_perform_step_action(self, every_nth_step: int, current_step: int):
        if every_nth_step < 0:
            return False
        last_step = current_step == self.sync_steps
        if every_nth_step == 0:
            return last_step
        else:
            return (current_step % every_nth_step == 0) or last_step


@dataclass
class EvalContext(BaseTrainContext):
    """Context used for evaluation"""

    @classmethod
    def from_train_context(cls: 'EvalContext', context: TrainContext):
        return cls(
            trainer=context.trainer,
            logger=context.logger,
            optimizer=context.optimizer,
            scheduler=context.scheduler,
            data_loader=context.evaluate_data_loader,
            model=context.model,
            model_unwrapped=context.model_unwrapped,
            train_state=context.train_state
        )

    @property
    def context_type(self) -> ContextType:
        return ContextType.EVAL


class InferContext(BaseContext):
    """Context used for inference"""

    @property
    def context_type(self) -> ContextType:
        return ContextType.INFERENCE



import abc
from abc import ABC, abstractmethod

from torch import Tensor
from torchmetrics import Metric

from xztrainer.context import BaseContext, ContextType, TrainContext, BaseTrainContext
from xztrainer.model import DataType, ModelOutputsType, TrackerConfigType


class XZTrainable(ABC):
    """
    Class representing a custom training logic.
    """

    @abstractmethod
    def step(
            self,
            context: BaseContext,
            data: DataType
    ) -> tuple[Tensor, ModelOutputsType]:
        """
        A function that:

        1. Performs a forward pass
        2. Computes loss function value
        3. Outputs loss value along with arbitrary model artifacts (used for metric computation further)

        Args:
            context: Current **train**/**evaluation** context.
            data: Model inputs from a collating function, moved to a target device

        Returns:
            A tuple containing (loss function value tensor, dictionary of model output artifacts)
        """

        ...

    def cut_model_outputs(
            self,
            context: BaseContext,
            model_outputs: ModelOutputsType,
            remainder: int
    ) -> ModelOutputsType:
        """
        Function used to cut out `model_outputs` based on a `remainder` value.

        Since accelerate repeats some last samples in both validation and training datasets
        (let's call these samples "padded samples") to ensure batches are split across GPUs evenly in distributed
        setups, you may want to cut out these padded samples before metric calculation.

        Default implementation just doesn't do anything because it does not know model output artifacts structure.

        If you doesn't implement this function, you may see a slight difference between calculated metrics for a distributed
        and non-distributed setups.

        Examples:
            >>> return {'predictions': model_outputs['predictions'][:-remainder]}

        Args:
            context: Current **train**/**evaluation** context
            model_outputs: Model output artifacts
            remainder: Number of last samples that are padded samples and need to be cut out.

        """
        return model_outputs

    @abc.abstractmethod
    def create_metrics(self, context_type: ContextType) -> dict[str, Metric]:
        """
        Creates a dictionary of {metric name: metric object}. Use metrics from `torchmetrics` package.
        You can create different metrics for different contexts (train, eval).

        Args:
            context_type: Current context type
        """
        ...

    @abc.abstractmethod
    def update_metrics(self, context_type: ContextType, model_outputs: ModelOutputsType, metrics: dict[str, Metric]):
        """
        Updates metric states after `step()` function using artifacts from it

        Args:
            context_type: Current context type
            model_outputs: Model output artifacts from `step()` call
            metrics: Metric objects from `create_metrics()` call
        """
        ...

    def calculate_composition_metrics(self, context_type: ContextType, metric_values: dict[str, float]) -> dict[str, float]:
        """
        Utility function to calculate metrics that depends on other metrics only.

        Args:
            context_type: Current context type
            metric_values: Calculated other metric values
        """
        return {}

    def on_load(self, context: TrainContext, step: int):
        """
        Callback that is called when a training is either started from scratch or started from a checkpoint.

        Use it, for example, to initialize weight tensors for entropy loss.

        Args:
            context: Current **train** context
            step: Current training step. Will be `1` when training is started from scratch. Will be other positive value when training is started from a checkponit.
        """
        pass

    def log(self, context: BaseTrainContext):
        """
        Callback that is called when trainer logs metrics values (both for training and evaluation).

        Args:
            context: Current **train**/**evaluation** context
        """
        pass

    def on_pre_update(self, context: TrainContext, step: int):
        """
        Callback that is called **before** an optimizer step occurs.

        Args:
            context: Current **train** context
            step: Current training step
        """
        pass

    def on_update(self, context: TrainContext, step: int):
        """
        Callback that is called **after** an optimizer step occurs.

        Args:
            context: Current **train** context
            step: Current training step
        """
        pass

    def tracker_config(self, context: TrainContext) -> TrackerConfigType:
        """
        Function returning additional hyperparameters that will be logged to experiment tracker. This function
        is called once when training starts.

        Args:
            context: Current **train** context

        Returns: Dictionary with additional parameter names and their values.

        """
        return {}



from typing import Union, Iterable

from accelerate import Accelerator

ClassifierType = Union[str, Iterable[str]]
"""A type that represents a classifier, string ("perplexity")  or collection of strings (["perplexity", "books"])"""


def _convert_classifier(classifier: ClassifierType) -> Iterable[str]:
    if isinstance(classifier, str):
        return classifier,
    elif isinstance(classifier, Iterable):
        return tuple(classifier)
    else:
        raise ValueError(f'Invalid classifier type: {type(classifier)}')


class AccelerateLogger:
    """
    A wrapper for Accelerate logging API that is able to update its current time step and top-level classifier
    (train/eval/etc...) by xztrainer loop. It can be used within
    """

    def __init__(self, accelerator: Accelerator):
        self._time_step = 0
        self._top_classifier = ()
        self._accelerator = accelerator

    def log_scalar(self, classifier: ClassifierType, value: float):
        """
        Logs a scalar value
        :param classifier: Value classifier (e.g. "loss" or "perplexity")
        :param value: Scalar value
        """
        classifier = '/'.join(self._top_classifier + _convert_classifier(classifier))
        step = self._time_step
        self._accelerator.log({classifier: value}, step=step)

    def update_time_step(self, time_step: int):
        """
        Sets current time step of this logger (used by xztrainer internals).
        :param time_step: Time step to set
        """
        self._time_step = time_step

    def update_top_classifier(self, classifier: ClassifierType):
        """
        Sets top-level classifier of this logger (used by xztrainer internals).
        :param classifier: Classifier to set
        """
        self._top_classifier = _convert_classifier(classifier)



from typing import Callable

from torch import nn


def count_parameters(module: nn.Module, parameter_predicate: Callable[[nn.Parameter], bool] = lambda p: True) -> int:
    """
    Filters all the module parameter tensors based on a given predicate and counts a total parameter count after that.
    :param module: Module to count parameters in
    :param parameter_predicate: Predicate to filter out model parameters to be counted
    :return: Parameter count
    """
    return sum(param.numel() for param in module.parameters() if parameter_predicate(param))
